
> module Parser (parse) where

> import Char
> import PC
> import Tree
> import Positionise

> pE :: Parser a b -> Parser a c -> Parser a (Either b c)
> pE p1 p2 = Left <$> p1  |> Right <$> p2

> parse :: [(Char, Position)] -> Tree
> parse inp = if parse_succeeded p && null (rest_of_input p)
>             then ret_val p
>             else error ("Parser failed at " ++ show_pos (fail_pos p))
>   where p = parse' inp

> parse' :: Parser Char Tree
> parse' = Tree <$> pList 0 (pE line def)

> line :: Parser Char Line
> line = Line
>    <$> pList 0 (pPred ('\n' /=))
>    <*  pSym '\n'

> def :: Parser Char Def
> def = (\s n w1 b w2 es d -> Def False n (s:n ++ w1 ++ b ++ w2 ++ concat es ++ d))
>   <$> pSym '/'
>   <*> pList 1 (pPred is_name_char)
>   <*> pWhite
>   <*> (pBalanced <|> pLit)
>   <*> pWhite
>   <*> pList 0 (    (\xs ss -> xs ++ ss)
>                <$> pList 0 (pPred (not . isSpace))
>                <*> pList 1 (pSym ' ')
>               )
>   <*> pSyms "def\n"

> pLit :: Parser Char String
> pLit = pList 1 (pPred (\c -> not (isSpace c) && not_brace c))

> is_name_char :: Char -> Bool
> is_name_char c = isAlphaNum c || c == '_' || c == '.'

> pWhite :: Parser Char String
> pWhite = (\l m -> case m of
>                       Nothing -> l
>                       Just s -> l ++ s)
>      <$> pList 0 (pPred isSpace)
>      <*> pMaybe (    (\p c n w -> [p] ++ c ++ [n] ++ w)
>                  <$> pSym '%'
>                  <*> pList 0 (pPred ('\n' /=))
>                  <*> pSym '\n'
>                  <*> pWhite
>                 )

> pBalanced :: Parser Char String
> pBalanced = (\co first rest cc -> [co] ++ first ++ concat rest ++ [cc])
>         <$> pSym '{'
>         <*> pList 0 (pPred not_brace)
>         <*> pList 0 (    (++)
>                      <$> pBalanced
>                      <*> pList 0 (pPred not_brace)
>                     )
>         <*> pSym '}'

> not_brace :: Char -> Bool
> not_brace '{' = False
> not_brace '}' = False
> not_brace _ = True

